from bppy.execution.listeners.b_program_runner_listener import BProgramRunnerListener


class PrintBProgramRunnerListener(BProgramRunnerListener):

    def starting(self, b_program):
        print("STARTED")

    def started(self, b_program):
        pass

    def super_step_done(self, b_program):
        pass

    def ended(self, b_program):
        print("ENDED")

    def assertion_failed(self, b_program):
        pass

    def b_thread_added(self, b_program):
        pass

    def b_thread_removed(self, b_program):
        pass

    def b_thread_done(self, b_program):
        pass

    def event_selected(self, b_program, event):
        print(event)

    def halted(self, b_program):
        pass

    def __init__(self):
        super().__init__()





from bppy.model.event_selection.event_selection_strategy import EventSelectionStrategy
from abc import abstractmethod

class SolverBasedEventSelectionStrategy(EventSelectionStrategy):
    @abstractmethod
    def select(self, statements, external_events_queue=[]):
        pass

    @abstractmethod
    def is_satisfied(self, event, statement):
        pass



from bppy.model.b_event import BEvent
from inspect import signature


class EventSet:

    def __init__(self, predicate, **kwargs):
        self.predicate = predicate
        self.data = kwargs

    def __contains__(self, event):
        if len(signature(self.predicate).parameters) > 1:
            return self.predicate(event, self.data)
        else:
            return self.predicate(event)


class All(EventSet):
    def __init__(self):
        super().__init__(lambda e: True)


class EmptyEventSet(EventSet):
    def __init__(self):
        super().__init__(lambda e: False)


class AllExcept(EventSet):
    def __init__(self, event):
        if isinstance(event, BEvent):
            super().__init__(lambda e: event != e)
        else:  # Event Set
            super().__init__(lambda e: not event.__contains__(e))


class EventSetList(EventSet):
    def __init__(self, lst):
        self.lst = lst
        super().__init__(lambda e: any([EventSetList.item_contains(item, e) for item in self.lst]))

    @staticmethod
    def item_contains(item, event):
        if isinstance(item, BEvent):
            return event == item
        elif len(signature(item.predicate).parameters) > 1:
            return item.predicate(event, item.data)
        else:
            return item.predicate(event)




from abc import ABC, abstractmethod


class BProgramRunnerListener(ABC):

    @abstractmethod
    def starting(self, b_program):
        pass

    @abstractmethod
    def started(self, b_program):
        pass

    @abstractmethod
    def super_step_done(self, b_program):
        pass

    @abstractmethod
    def ended(self, b_program):
        pass

    @abstractmethod
    def assertion_failed(self, b_program):
        pass

    @abstractmethod
    def b_thread_added(self, b_program):
        pass

    @abstractmethod
    def b_thread_removed(self, b_program):
        pass

    @abstractmethod
    def b_thread_done(self, b_program):
        pass

    @abstractmethod
    def event_selected(self, b_program, event):
        pass

    @abstractmethod
    def halted(self, b_program):
        pass




from importlib import import_module
from inspect import getmembers, isfunction
from itertools import tee
import warnings

from z3 import *


class BProgram:

    def __init__(self, bthreads=None, source_name=None, event_selection_strategy=None, listener=None):
        self.source_name = source_name
        self.new_bt = []
        self.bthreads = bthreads
        self.event_selection_strategy = event_selection_strategy
        self.listener = listener
        self.variables = None
        self.tickets = []
        self.external_events_queue = []

    def setup(self):
        if self.source_name:
            self.bthreads = [o[1]() for o in getmembers(import_module(self.source_name)) if
                             isfunction(o[1]) and o[1].__module__ == self.source_name]

            self.variables = dict([o for o in getmembers(import_module(self.source_name)) if
                                   isinstance(o[1], ExprRef) or isinstance(o[1], list)])

        self.new_bt = self.bthreads
        self.load_new_bthreads()


    def advance_bthreads(self,tickets, m):
        if len(self.new_bt) > 0:
            warnings.warn("Some new bthreads are not loaded and are not affecting the bprogram. Use the load_new_bthreads method to load them.", RuntimeWarning)
        for l in tickets:
            if m is None or self.event_selection_strategy.is_satisfied(m, l):
                try:
                    bt = l['bt']
                    l.clear()
                    ll = bt.send(m)
                    if ll is None:
                        continue
                    l.update(ll)
                    l.update({'bt': bt})
                except (KeyError, StopIteration):
                    pass

    def add_bthread(self, bt):
        self.new_bt.append(bt)

    def load_new_bthreads(self):
        while len(self.new_bt) > 0:
            new_tickets = [{'bt': bt} for bt in self.new_bt]
            self.new_bt.clear()
            self.advance_bthreads(new_tickets, None)
            self.tickets.extend(new_tickets)

    def next_event(self):
        if len(self.new_bt) > 0:
            warnings.warn(
                "Some new bthreads are not loaded and are not affecting the bprogram. Use the load_new_bthreads method to load them.",
                RuntimeWarning)
        return self.event_selection_strategy.select(self.tickets, self.external_events_queue)

    def run(self):
        if self.listener:
            self.listener.starting(b_program=self)

        self.setup()
        # Main loop
        interrupted = False
        while not interrupted:
            # for dynamically added bthreads
            self.load_new_bthreads()

            event = self.next_event()
            # Finish the program if no event is selected
            if event is None:
                break
            if self.listener:
                interrupted = self.listener.event_selected(b_program=self, event=event)

            self.advance_bthreads(self.tickets,event)


        if self.listener:
            self.listener.ended(b_program=self)

    def enqueue_external_event(self, event):
        self.external_events_queue.append(event)



from bppy.model.event_selection.event_selection_strategy import EventSelectionStrategy
from bppy.model.b_event import BEvent
from bppy.model.event_set import EmptyEventSet
import random
from collections.abc import Iterable


class SimpleEventSelectionStrategy(EventSelectionStrategy):

    def is_satisfied(self, event, statement):
        if isinstance(statement.get('request'), BEvent):
            if isinstance(statement.get('waitFor'), BEvent):
                return statement.get('request') == event or statement.get('waitFor') == event
            else:
                return statement.get('request') == event or statement.get('waitFor', EmptyEventSet()).__contains__(event)
        else:
            if isinstance(statement.get('waitFor'), BEvent):
                return statement.get('request', EmptyEventSet()).__contains__(event) or statement.get('waitFor') == event
            else:
                return statement.get('request', EmptyEventSet()).__contains__(event) or statement.get('waitFor', EmptyEventSet()).__contains__(event)

    def selectable_events(self, statements):
        possible_events = set()
        for statement in statements:
            if 'request' in statement:  # should be eligible for sets
                if isinstance(statement['request'], Iterable):
                    possible_events.update(statement['request'])
                elif isinstance(statement['request'], BEvent):
                    possible_events.add(statement['request'])
                else:
                    raise TypeError("request parameter should be BEvent or iterable")
        for statement in statements:
            if 'block' in statement:
                if isinstance(statement.get('block'), BEvent):
                    possible_events.discard(statement.get('block'))
                else:
                    possible_events = {x for x in possible_events if x not in statement.get('block')}
        return possible_events

    def select(self, statements, external_events_queue=[]):
        selectable_events = self.selectable_events(statements)
        if selectable_events:
            return random.choice(tuple(selectable_events))
        else:
            if len(external_events_queue) > 0:
                return external_events_queue.pop(0)
            else:
                return None




from abc import ABC, abstractmethod


class EventSelectionStrategy(ABC):

    @abstractmethod
    def select(self, statements, external_events_queue=[]):
        pass

    @abstractmethod
    def is_satisfied(self, event, statement):
        pass
